@@ -39,7 +39,9 @@
 
 #define KB			(1024ULL)
 #define MB			(KB * KB)
-#define GB			(KB * KB * KB)
+#define GB			(KB * MB)
+#define TB			(KB * GB)
+#define PB			(KB * TB)
 
 #define UNDERRUN_MAX		(100)		/* Max underruns before warning, see -w */
 #define UNDERRUN_ADJUST_MAX	(1)		/* Underruns before adjusting rate */
@@ -49,7 +51,7 @@
 #define DELAY_SHIFT_MAX		(16)		/* Max shift, see -s */
 
 #define IO_SIZE_MIN		(1)		/* Min io buffer size, see -i */
-#define IO_SIZE_MAX		(MB * 64)	/* Max io buffer size, see -i */
+#define IO_SIZE_MAX		(1ULL * GB)	/* Max io buffer size, see -i */
 
 #define DELAY_MIN		(0.01)		/* Min delay time, see -c */
 #define DELAY_MAX		(10.00)		/* Max delay time, see -c */
@@ -342,7 +344,7 @@ static const char *double_to_str(const double val)
  */
 static void stats_info(const stats_t *stats)
 {
-	double secs = stats->time_end - stats->time_begin;
+	const double secs = stats->time_end - stats->time_begin;
 	double avg_wr_sz;
 	struct tms t;
 
@@ -476,13 +478,13 @@ static uint64_t get_uint64(const char *const str, size_t *const len)
  *  get_double()
  *	get a double value
  */
-static double get_double(const char *const str, size_t *const len)
+static double get_double(const char *const str, char **endptr, size_t *const len)
 {
 	double val;
 	*len = strlen(str);
 
 	errno = 0;
-	val = strtod(str, NULL);
+	val = strtod(str, endptr);
 	if (errno) {
 		fprintf(stderr, "Invalid value %s.\n", str);
 		exit(EXIT_BAD_OPTION);
@@ -507,14 +509,26 @@ static double get_double_scale(
 	double val;
 	size_t len = strlen(str);
 	int i;
-	char ch;
+	char ch, *endptr;
 
-	val = get_double(str, &len);
-	len--;
-	ch = str[len];
+	endptr = NULL;
+	val = get_double(str, &endptr, &len);
+
+	if (!endptr)
+		return val;
+
+	ch = *endptr;
+	if (!ch)
+		return val;
+
+	if (*(endptr + 1)) {
+		fprintf(stderr, "Expecting 1 character size specifier, got '%s'.\n",
+			endptr);
+		exit(EXIT_BAD_OPTION);
+	}
 
 	if (val < 0.0) {
-		printf("Value %s cannot be negative\n", str);
+		fprintf(stderr, "Value %s cannot be negative\n", str);
 		exit(EXIT_BAD_OPTION);
 	}
 
@@ -527,7 +541,7 @@ static double get_double_scale(
 			return val * scales[i].scale;
 	}
 
-	printf("Illegal %s specifier %c\n", msg, str[len]);
+	fprintf(stderr, "Illegal %s specifier '%c'\n", msg, *endptr);
 	exit(EXIT_BAD_OPTION);
 }
 
@@ -580,7 +594,7 @@ static void show_usage(void)
 	printf("Usage: %s [options]\n", app_name);
 	printf("  -a         append to file (-t, -O options only).\n");
 	printf("  -c delay   specify constant delay time (seconds).\n");
-	printf("  -d         discard input (no output).\n");
+	printf("  -d         discard output (no output).\n");
 	printf("  -D         delay mode.\n");
 	printf("  -e         skip read errors.\n");
 	printf("  -f freq    frequency of -v statistics.\n");
@@ -591,7 +605,7 @@ static void show_usage(void)
 	printf("  -o         shrink read/write buffer to avoid overrun.\n");
 	printf("  -O file    short cut for -dt file; output to a file.\n");
 	printf("  -p         enable verbose mode with progress stats.\n");
-	printf("  -P pidfile save process ID intp file pidfile.\n");
+	printf("  -P pidfile save process ID into file pidfile.\n");
 	printf("  -r rate    set rate (in bytes per second).\n");
 	printf("  -R	     ignore stdin, read from %s.\n", dev_urandom);
 	printf("  -s shift   controls delay or buffer size adjustment.\n");
@@ -755,6 +769,10 @@ int main(int argc, char **argv)
 			break;
 		case 'r':
 			data_rate = get_double_byte(optarg);
+			if (data_rate > 1.0 * PB) {
+				fprintf(stderr, "Data rate too high.\n");
+				exit(EXIT_BAD_OPTION);
+			}
 			opt_flags |= OPT_GOT_RATE;
 			break;
 		case 'R':
@@ -828,7 +846,7 @@ int main(int argc, char **argv)
 		goto tidy;
 	}
 	if (!(opt_flags & (OPT_GOT_RATE | OPT_NO_RATE_CONTROL))) {
-		fprintf(stderr, "Must specify data rate with -r option.\n");
+		fprintf(stderr, "Must specify data rate with -r option (or use -n for no rate control).\n");
 		ret = EXIT_BAD_OPTION;
 		goto tidy;
 	}
@@ -899,11 +917,14 @@ int main(int argc, char **argv)
 				if (io_size < IO_SIZE_MIN)
 					io_size = IO_SIZE_MIN;
 				if (io_size > IO_SIZE_MAX) {
-					fprintf(stderr, "Rate too high, maximum allowed: %s/sec.\n",
+					io_size = IO_SIZE_MAX;
+					/*
+					fprintf(stderr, "Rate too high for the buffer size, maximum allowed: %s/sec.\n",
 						double_to_str((double)IO_SIZE_MAX * 32.0));
-					fprintf(stderr, "Use -i to explicitly set buffer size.\n");
+					fprintf(stderr, "Use -i to explicitly set a larger buffer size.\n");
 					ret = EXIT_BAD_OPTION;
 					goto tidy;
+					*/
 				}
 			}
 		}
